from _thread import start_new_thread, allocate_lock
import json
import zlib
import ssl
import uuid
import _socket
import xbmc
import xbmcgui
from helper import utils, queue, artworkcache
from database import dbio
from core import common


class HTTP:
    def __init__(self, EmbyServer):
        self.EmbyServer = EmbyServer
        self.Intros = []
        self.RequestHeader = {}
        self.Queues = {"ASYNC": queue.Queue(), "DOWNLOAD": queue.Queue()}
        self.HostName = ""
        self.Port = 0
        self.socket = {}
        self.SocketBusy = {"MAIN": allocate_lock()}

    def socket_open(self, ConnectionString, SocketId):
        Scheme, Hostname, Port = utils.get_url_info(ConnectionString)
        self.RequestHeader[SocketId] = {"Host": f"{Hostname}:{Port}", 'Content-type': 'application/json; charset=utf-8', 'Accept-Charset': 'utf-8', 'Accept-encoding': 'gzip', 'User-Agent': f"{utils.addon_name}/{utils.addon_version}", 'Connection': 'keep-alive', 'Authorization': f'Emby Client="{utils.addon_name}", Device="{utils.device_name}", DeviceId="{self.EmbyServer.ServerData["DeviceId"]}", Version="{utils.addon_version}"'}

        try:
            address_family = _socket.getaddrinfo(Hostname, None)[0][0]
        except Exception as error:
            xbmc.log(f"EMBY.emby.http: Socket open {SocketId}: Wrong Hostname {error}", 2) # LOGWARNING
            utils.Dialog.notification(heading=utils.addon_name, icon="DefaultIconError.png", message="Invalid server address", time=utils.displayMessage, sound=False)
            return 609

        TimeoutCounter = 0

        while True:
            self.socket[SocketId] = None

            try:
                self.socket[SocketId] = _socket.socket(address_family, _socket.SOCK_STREAM)
                self.socket[SocketId].setsockopt(_socket.IPPROTO_TCP, _socket.TCP_NODELAY, 1)
                self.socket[SocketId].settimeout(1) # set timeout
                self.socket[SocketId].connect((Hostname, Port))
                break
            except TimeoutError:
                TimeoutCounter += 1

                if TimeoutCounter < 10:
                    continue

                xbmc.log(f"EMBY.emby.http: Socket open {SocketId}: Timeout", 2) # LOGWARNING
                del self.socket[SocketId]
                return 606
            except ConnectionRefusedError:
                del self.socket[SocketId]
                xbmc.log(f"EMBY.emby.http: [ ServerUnreachable ] {SocketId}", 2) # LOGWARNING
                xbmc.log(f"EMBY.emby.http: [ ServerUnreachable ] {ConnectionString}", 0) # LOGDEBUG
                return 607
            except Exception as error:
                if str(error) == "timed out": # workaround when TimeoutError not raised
                    TimeoutCounter += 1

                    if TimeoutCounter < 10:
                        continue

                    xbmc.log(f"EMBY.emby.http: Socket open {SocketId}: Timeout", 2) # LOGWARNING
                    del self.socket[SocketId]
                    return 606

                if str(error).lower().find("errno 22") != -1 or str(error).lower().find("invalid argument") != -1: # [Errno 22] Invalid argument
                    del self.socket[SocketId]
                    xbmc.log(f"EMBY.emby.http: Socket open {SocketId}: Invalid argument", 2) # LOGWARNING
                    utils.Dialog.notification(heading=utils.addon_name, icon="DefaultIconError.png", message="Invalid argument", time=utils.displayMessage, sound=False)
                    return 610

                xbmc.log(f"EMBY.emby.http: Socket open {SocketId}: Undefined error: {error}", 2) # LOGWARNING
                xbmc.log(f"EMBY.emby.http: Socket open {SocketId}: Undefined error type {type(error)}", 2) # LOGWARING
                return 699

        self.HostName = Hostname
        self.Port = Port

        if Scheme == "https":
            SSLContext = ssl.SSLContext(ssl.PROTOCOL_TLS)
            SSLContext.load_default_certs()

            if utils.sslverify:
                SSLContext.verify_mode = ssl.CERT_REQUIRED
            else:
                SSLContext.verify_mode = ssl.CERT_NONE

            try:
                self.socket[SocketId] = SSLContext.wrap_socket(self.socket[SocketId], do_handshake_on_connect=True, suppress_ragged_eofs=True, server_hostname=Hostname)
            except ssl.CertificateError:
                del self.socket[SocketId]
                xbmc.log("EMBY.emby.http: socket_open ssl certificate error", 3) # LOGERROR
                utils.Dialog.notification(heading=utils.addon_name, message=utils.Translate(33428), time=utils.displayMessage)
                return 608
            except Exception as error:
                del self.socket[SocketId]
                xbmc.log(f"EMBY.emby.http: socket_open ssl undefined error: {error}", 2) # LOGWARNING
                return 699

        if SocketId == "MAIN":
            self.Queues["ASYNC"].clear()
            self.Queues["DOWNLOAD"].clear()
            start_new_thread(self.Ping, ())
            start_new_thread(self.async_commands, ())
            start_new_thread(self.download_file, ())

        xbmc.log(f"EMBY.emby.http: Socket {SocketId} opened", 0) # LOGDEBUG
        return 0

    def socket_open_async(self, Id):
        if Id not in self.socket:
            xbmc.log(f"EMBY.emby.http: {Id} open socket connection", 1) # LOGINFO
            StatusCode = self.socket_open(self.EmbyServer.ServerData['ServerUrl'], Id)

            if StatusCode:
                xbmc.log(f"EMBY.emby.emby: THREAD: ---<[ {Id} {self.EmbyServer.ServerData['ServerId']} Connection failed: {StatusCode} ]", 2) # LOGWARNING
                return StatusCode

            # Update Header information
            if 'X-Emby-Token' not in self.RequestHeader[Id] and  self.EmbyServer.ServerData['AccessToken'] and self.EmbyServer.ServerData['UserId']:
                self.RequestHeader[Id].update({'Authorization': f'{self.RequestHeader[Id]["Authorization"]}, Emby UserId="{self.EmbyServer.ServerData["UserId"]}"', 'X-Emby-Token': self.EmbyServer.ServerData['AccessToken']})

        return 0

    def socket_close(self, SocketId):
        if SocketId == "ALL":
            SocketIdsLocal = list(self.socket.keys())
        else:
            SocketIdsLocal = [SocketId]

        for SocketIdLocal in SocketIdsLocal:
            if SocketIdLocal not in self.socket:
                xbmc.log(f"EMBY.emby.http: Socket {SocketIdLocal} already closed", 0) # LOGDEBUG
                continue

            if SocketIdLocal == "MAIN":
                self.Queues["ASYNC"].put("QUIT")
                self.Queues["DOWNLOAD"].put("QUIT")

            if SocketIdLocal in self.socket:
                try:
                    self.socket[SocketIdLocal].close()
                except Exception as error:
                    xbmc.log(f"EMBY.emby.http: Socket {SocketIdLocal} close error: {error}", 2) # LOGWARNING

                try:
                    del self.socket[SocketIdLocal]
                    del self.RequestHeader[SocketIdLocal]
                except Exception as error:
                    xbmc.log(f"EMBY.emby.http: Socket {SocketIdLocal} reset error: {error}", 2) # LOGWARNING

            if SocketIdLocal != "MAIN" and SocketIdLocal in self.SocketBusy:
                del self.SocketBusy[SocketIdLocal]

            xbmc.log(f"EMBY.emby.http: Socket {SocketIdLocal} closed", 0) # LOGDEBUG

    def socket_io(self, Request, SocketId, Timeout):
        IncomingData = b""
        StatusCode = 0
        TimeoutCounter = 0
        WaitCounter = 0
        BytesSend = 0
        BytesSendTotal = len(Request)

        while True:
            try:
                self.socket[SocketId].settimeout(1) # set timeout

                if Request:
                    while utils.PlayerBusy and WaitCounter < 30: # Timeout after 30 seconds
                        if not WaitCounter:
                            xbmc.log(f"EMBY.emby.http: Socket IO {SocketId}: THREAD: --->[ Player busy ]", 1) # LOGDEBUG

                        if utils.sleep(1):
                            xbmc.log(f"EMBY.emby.http: Socket IO {SocketId}: THREAD: ---<[ Player busy ] Shutdown", 1) # LOGDEBUG
                            return 601, b""

                        if not WaitCounter:
                            xbmc.log(f"EMBY.emby.http: Socket IO {SocketId}: THREAD: ---<[ Player busy ]", 1) # LOGDEBUG

                        WaitCounter += 1

                    while BytesSend < BytesSendTotal:
                        BytesSend += self.socket[SocketId].send(Request[BytesSend:])
                else:
                    IncomingData = self.socket[SocketId].recv(1048576)

                    if not IncomingData: # No Data received
                        xbmc.log(f"EMBY.emby.http: Socket IO {SocketId}: ({bool(Request)}): Empty data", 2) # LOGWARNING
                        StatusCode = 600

                break
            except TimeoutError:
                if SocketId != "MAIN" and self.SocketBusy["MAIN"].locked(): # Wait longer for e.g. images. MAIN queries could block IO
                    continue

                TimeoutCounter += 1

                if TimeoutCounter < Timeout:
                    continue

                xbmc.log(f"EMBY.emby.http: Socket IO {SocketId}: ({bool(Request)}): Timeout", 2) # LOGWARNING
                StatusCode = 603
                break
            except BrokenPipeError:
                xbmc.log(f"EMBY.emby.http: Socket IO {SocketId}: ({bool(Request)}): Pipe error", 2) # LOGWARNING
                StatusCode = 605
                break
            except Exception as error:
                if str(error) == "timed out": # workaround when TimeoutError not raised
                    if SocketId != "MAIN" and self.SocketBusy["MAIN"].locked(): # Wait longer for e.g. images. MAIN queries could block IO
                        continue

                    TimeoutCounter += 1

                    if TimeoutCounter < Timeout:
                        continue

                    xbmc.log(f"EMBY.emby.http: Socket IO {SocketId}: ({bool(Request)}): Timeout (workaround)", 2) # LOGWARNING
                    StatusCode = 603
                    break

                xbmc.log(f"EMBY.emby.http: Socket IO {SocketId}: ({bool(Request)}): Undefined error {error}", 3) # LOGERROR
                xbmc.log(f"EMBY.emby.http: Socket IO {SocketId}: ({bool(Request)}): Undefined error type {type(error)}", 3) # LOGERROR
                StatusCode = 699
                break

        return StatusCode, IncomingData

    def socket_request(self, Method, Handler, Params, Binary, TimeoutSend, TimeoutRecv, SocketId, DownloadPath, DownloadName):
        if SocketId not in self.socket:
            return 601, {}, {}

        PayloadTotal = b""
        PayloadTotalPosition = 0

        # Prepare HTTP Header
        HeaderString = ""

        for Key, Values in list(self.RequestHeader[SocketId].items()):
            HeaderString += f"{Key}: {Values}\r\n"

        # Prepare HTTP Payload
        if Method == "GET":
            ParamsString = ""

            for Query, Param in list(Params.items()):
                if Param not in ([], None):
                    ParamsString += f"{Query}={Param}&"

            if ParamsString:
                ParamsString = f"?{ParamsString[:-1]}"

            StatusCodeSocket, _ = self.socket_io(f"{Method} /{Handler}{ParamsString} HTTP/1.1\r\n{HeaderString}Content-Length: 0\r\n\r\n".encode("utf-8"), SocketId, TimeoutSend)
        else:
            if Params:
                ParamsString = json.dumps(Params)
            else:
                ParamsString = ""

            StatusCodeSocket, _ = self.socket_io(f"{Method} /{Handler} HTTP/1.1\r\n{HeaderString}Content-Length: {len(ParamsString)}\r\n\r\n{ParamsString}".encode("utf-8"), SocketId, TimeoutSend)

        if StatusCodeSocket:
            return StatusCodeSocket, {}, ""

        IncomingData = b""

        # Receive Data
        while True:
            StatusCodeSocket, IncomingData = self.socket_io("", SocketId, TimeoutRecv)

            if StatusCodeSocket:
                return StatusCodeSocket, {}, ""

            IncomingData = IncomingData.split(b'\x0d\x0a\x0d\x0a', 1) # Receive initial package 1MB
            IncomingDataHeadArray = IncomingData[0].decode("utf-8").split("\r\n")
            StatusCode = int(IncomingDataHeadArray[0].split(" ")[1])

            if StatusCode not in (200, 206) or StatusCode == 204: # 200 = OK, 206 = Partial content, 204 = OK but no content
                return StatusCode, {}, ""

            IncomingDataHeaderArray = IncomingDataHeadArray[1:]
            IncomingDataHeader = {}

            for IncomingDataHeaderArrayData in IncomingDataHeaderArray:
                Temp = IncomingDataHeaderArrayData.split(": ")
                IncomingDataHeader[Temp[0]] = Temp[1]

            isGzip = IncomingDataHeader.get("Content-Encoding", "") == "gzip"
            isDeflate = IncomingDataHeader.get("Content-Encoding", "") == "deflate"

            # Recv frame
            PayloadLenghtTotal = int(IncomingDataHeader["Content-Length"])
            PayloadTotalPosition += PayloadLenghtTotal
            Payload = IncomingData[1]
            del IncomingData
            PayloadLenght = len(Payload)

            if DownloadPath:
                ProgressBar = xbmcgui.DialogProgressBG()
                ProgressBar.create("Download", DownloadName)
                ProgressBarTotal = PayloadLenghtTotal / 100
                OutFile = open(DownloadPath, 'wb')

            while PayloadLenght < PayloadLenghtTotal:
                StatusCodeSocket, PayloadRecv = self.socket_io("", SocketId, TimeoutRecv)

                if StatusCodeSocket:
                    if DownloadPath:
                        OutFile.close()
                        ProgressBar.close()
                        del ProgressBar

                    return StatusCodeSocket, {}, ""

                Payload += PayloadRecv
                PayloadLenght += len(PayloadRecv)
                del PayloadRecv

                if DownloadPath:
                    # Decompress data
                    if isDeflate:
                        Payload = zlib.decompress(PayloadTotal, zlib.MAX_WBITS|32)
                    elif isGzip:
                        Payload = zlib.decompress(Payload, zlib.MAX_WBITS|32)

                    OutFile.write(Payload)
                    Payload = b""
                    ProgressBar.update(int(PayloadLenght / ProgressBarTotal), "Download", DownloadName)

            if DownloadPath:
                OutFile.close()
                ProgressBar.close()
                del ProgressBar

            PayloadTotal += Payload
            del Payload

            # request additional data
            if StatusCode == 206: # partial content
                xbmc.log(f"EMBY.emby.http: Partial content {SocketId} closed", 1) # LOGINFO

                if Method == "GET":
                    StatusCodeSocket, _ = self.socket_io(f"{Method} /{Handler}{ParamsString} HTTP/1.1\r\n{HeaderString}Range: bytes={PayloadTotalPosition}-\r\nContent-Length: 0\r\n\r\n".encode("utf-8"), SocketId, TimeoutSend)
                else:
                    StatusCodeSocket, _ = self.socket_io(f"{Method} /{Handler} HTTP/1.1\r\n{HeaderString}Content-Length: {len(ParamsString)}\r\n\r\n{ParamsString}".encode("utf-8"), SocketId, TimeoutSend)

                if StatusCodeSocket:
                    return StatusCodeSocket, {}, ""
            else:
                break

        # Decompress data
        if isDeflate:
            PayloadTotal = zlib.decompress(PayloadTotal, -zlib.MAX_WBITS)
        elif isGzip:
            PayloadTotal = zlib.decompress(PayloadTotal, zlib.MAX_WBITS|32)

        if Binary:
            if DownloadPath and PayloadTotal:
                with open(DownloadPath, 'wb') as OutFile:
                    OutFile.write(PayloadTotal)

                PayloadTotal = b""

            return StatusCode, IncomingDataHeader, PayloadTotal

        return StatusCode, IncomingDataHeader, json.loads(PayloadTotal)

    def download_file(self):
        xbmc.log("EMBY.emby.http: THREAD: --->[ file download ]", 0) # LOGDEBUG

        while True:
            Command = self.Queues["DOWNLOAD"].get() # EmbyId, ParentPath, Path, FilePath, FileSize, Name, KodiType, KodiPathIdBeforeDownload, KodiFileId, KodiId

            if Command == "QUIT":
                xbmc.log(f"EMBY.emby.emby: THREAD: ---<[ Download {self.EmbyServer.ServerData['ServerId']} ] shutdown 1", 0) # LOGDEBUG
                self.Queues["DOWNLOAD"].clear()
                utils.delFile(Command[3])
                self.socket_close("DOWNLOAD")
                return

            # check if free space below 2GB
            if utils.getFreeSpace(Command[2]) < (2097152 + Command[4] / 1024):
                utils.Dialog.notification(heading=utils.addon_name, message=utils.Translate(33429), icon=utils.icon, time=utils.displayMessage, sound=True)
                xbmc.log("EMBY.emby.http: THREAD: ---<[ file download ] terminated by filesize", 2) # LOGWARNING
                return

            while True:
                if self.socket_open_async("DOWNLOAD"):
                    if utils.sleep(10):
                        xbmc.log(f"EMBY.emby.emby: THREAD: ---<[ Download {self.EmbyServer.ServerData['ServerId']} shutdown ]", 0) # LOGDEBUG
                        return

                    continue

                StatusCode, _, _ = self.socket_request("GET", f"Items/{Command[0]}/Download", {}, True, 5, 10, "DOWNLOAD", Command[3], Command[5])

                if StatusCode == 601: # quit
                    xbmc.log(f"EMBY.emby.emby: THREAD: ---<[ Download {self.EmbyServer.ServerData['ServerId']} ] shutdown 2", 0) # LOGDEBUG
                    self.Queues["DOWNLOAD"].clear()
                    utils.delFile(Command[3])
                    self.socket_close("DOWNLOAD")
                    return

                if StatusCode in (600, 602, 603, 604, 605):
                    xbmc.log(f"EMBY.emby.http: Download retry {StatusCode}", 2) # LOGWARNING
                    utils.delFile(Command[3])
                    self.socket_close("DOWNLOAD")
                    continue

                try:
                    if StatusCode != 200:
                        utils.delFile(Command[3])
                    else:
                        if Command[9]: # KodiId
                            SQLs = dbio.DBOpenRW(self.EmbyServer.ServerData['ServerId'], "download_item", {})
                            SQLs['emby'].add_DownloadItem(Command[0], Command[7], Command[8], Command[9], Command[6])
                            dbio.DBCloseRW(self.EmbyServer.ServerData['ServerId'], "download_item", {})
                            SQLs = dbio.DBOpenRW("video", "download_item_replace", {})
                            Artworks = ()
                            ArtworksData = SQLs['video'].get_artworks(Command[9], Command[6])

                            for ArtworkData in ArtworksData:
                                if ArtworkData[3] in ("poster", "thumb", "landscape"):
                                    UrlMod = ArtworkData[4].split("|")
                                    UrlMod = f"{UrlMod[0].replace('-download', '')}-download|redirect-limit=1000"
                                    SQLs['video'].update_artwork(ArtworkData[0], UrlMod)
                                    Artworks += ((UrlMod,),)

                            SQLs['video'].update_Name(Command[9], Command[6], True)
                            SQLs['video'].replace_Path_ContentItem(Command[9], Command[6], Command[2])

                            if Command[6] == "episode":
                                KodiPathId = SQLs['video'].get_add_path(Command[2], None, Command[1])
                                Artworks = SQLs['video'].set_Subcontent_download_tags(Command[9], True)

                                if Artworks:
                                    artworkcache.CacheAllEntries(Artworks, None)
                            elif Command[6] == "movie":
                                KodiPathId = SQLs['video'].get_add_path(Command[2], "movie", None)
                            elif Command[6] == "musicvideo":
                                KodiPathId = SQLs['video'].get_add_path(Command[2], "musicvideos", None)
                            else:
                                KodiPathId = None
                                xbmc.log(f"EMBY.emby.http: Download invalid: KodiPathId: {Command[1]['Path']} / {Command[6]}", 2) # LOGWARNING

                            if KodiPathId:
                                SQLs['video'].replace_PathId(Command[8], KodiPathId)

                            dbio.DBCloseRW("video", "download_item_replace", {})
                            artworkcache.CacheAllEntries(Artworks, None)

                    if self.Queues["DOWNLOAD"].isEmpty():
                        utils.refresh_widgets(True)
                except Exception as error:
                    xbmc.log(f"EMBY.emby.http: Download Emby server did not respond: error: {error}", 2) # LOGWARNING

                break

        xbmc.log("EMBY.emby.http: THREAD: ---<[ file download ]", 0) # LOGDEBUG

    # decide threaded or wait for response
    def request(self, Method, Handler, Params, Header, Binary, ConnectionString, NewSocket):
        xbmc.log(f"EMBY.emby.http: [ http ] Method: {Method} / Handler: {Handler} / Params: {Params} / Binary: {Binary} / ConnectionString: {ConnectionString} / NewSocket: {NewSocket} / Header: {Header}", 0) # LOGDEBUG

        if NewSocket:
            SocketId = str(uuid.uuid4())
        else:
            SocketId = "MAIN"

        if SocketId not in self.SocketBusy:
            self.SocketBusy[SocketId] = allocate_lock()

        with self.SocketBusy[SocketId]:
            if not ConnectionString:
                ConnectionString = self.EmbyServer.ServerData['ServerUrl']

                if not NewSocket and not self.HostName or self.EmbyServer.ServerData["ServerUrl"].find(self.HostName) == -1:
                    self.socket_close(SocketId)
            else:
                if not NewSocket and not self.HostName or ConnectionString.find(self.HostName) == -1:
                    self.socket_close(SocketId)

            while True:
                StatusCode = 0

                # Shutdown
                if utils.SystemShutdown:
                    self.socket_close(SocketId)
                    return noData(StatusCode, Header, Binary)

                # open socket
                if SocketId not in self.socket:
                    StatusCode = self.socket_open(ConnectionString, SocketId)

                    if StatusCode:
                        if StatusCode not in (608, 609, 610): # wrong Emby server address or SSL issue
                            utils.Dialog.notification(heading=utils.addon_name, icon="DefaultIconError.png", message=utils.Translate(33575), time=utils.displayMessage, sound=False)
                            self.EmbyServer.ServerReconnect()

                        return noData(StatusCode, Header, Binary)

                # Update Header information
                if Header:
                    self.RequestHeader[SocketId] = {"Host": f"{self.HostName}:{self.Port}", 'Accept': "application/json", 'Accept-Charset': "utf-8", 'X-Application': f"{utils.addon_name}/{utils.addon_version}", 'Content-type': 'application/json'}
                    self.RequestHeader[SocketId].update(Header)
                elif 'X-Emby-Token' not in self.RequestHeader[SocketId] and self.EmbyServer.ServerData['AccessToken'] and self.EmbyServer.ServerData['UserId']:
                    self.RequestHeader[SocketId].update({'Authorization': f'{self.RequestHeader[SocketId]["Authorization"]}, Emby UserId="{self.EmbyServer.ServerData["UserId"]}"', 'X-Emby-Token': self.EmbyServer.ServerData['AccessToken']})

                StatusCode, Header, Payload = self.socket_request(Method, Handler, Params, Binary, 10, 300, SocketId, "", "")

                if NewSocket:
                    self.socket_close(SocketId)

                if StatusCode == 200: # OK
                    return StatusCode, Header, Payload

                if StatusCode == 204: # OK, no data
                    return noData(StatusCode, Header, Binary)

                if StatusCode == 401: # Unauthorized
                    utils.Dialog.notification(heading=utils.addon_name, message=utils.Translate(33147), time=utils.displayMessage)
                    return noData(StatusCode, Header, Binary)

                if StatusCode in (600, 604, 605): # not data received, broken pipes
                    xbmc.log(f"EMBY.emby.http: Request retry {StatusCode} / {SocketId}", 2) # LOGWARNING
                    self.socket_close(SocketId)
                    continue

                if StatusCode in (602, 603): # timeouts
                    self.socket_close(SocketId)
                    return noData(StatusCode, Header, Binary)

                if StatusCode == 601: # quit
                    return noData(StatusCode, Header, Binary)

                xbmc.log(f"EMBY.emby.http: [ Statuscode ] {StatusCode}", 3) # LOGERROR
                xbmc.log(f"EMBY.emby.http: [ Statuscode ] {Payload}", 0) # LOGDEBUG
                return noData(StatusCode, Header, Binary)

    # Delay low priority tasks
    def wait_for_main(self, Id):
        while self.SocketBusy["MAIN"].locked():
            if "MAIN" not in self.socket:
                xbmc.log(f"EMBY.emby.http: {Id} MAIN not running", 1) # LOGINFO

                if Id in self.Queues:
                    self.Queues[Id].clear()

                self.socket_close(Id)
                xbmc.log(f"EMBY.emby.emby: THREAD: ---<[ {Id} {self.EmbyServer.ServerData['ServerId']} MAIN socket closed]", 1) # LOGDEBUG
                return True

            if utils.sleep(2):
                xbmc.log(f"EMBY.emby.emby: THREAD: ---<[ {Id} {self.EmbyServer.ServerData['ServerId']} shutdown ]", 0) # LOGDEBUG
                self.socket_close(Id)
                return True

        return False

    # No return values are expected, usually also lower priority
    def async_commands(self):
        xbmc.log("EMBY.emby.http: THREAD: --->[ Async ]", 0) # LOGDEBUG
        CommandsTotal = ()

        # Sort tasks: priority tasks first, quit tasks second, others last
        while True:
            CommandsSorted = ()
            CommandsPriority = ()
            CommandsRegular = ()

            # merge commands
            Commands = self.Queues["ASYNC"].getall() # (Method, URL-handler, Parameters, Priority)
            CommandsTotal += Commands
            utils.sleep(0.25)

            if not self.Queues["ASYNC"].isEmpty():
                continue

            # Sort commands
            QUIT = False

            for CommandTotal in CommandsTotal:
                if len(CommandTotal) > 1:
                    if CommandTotal[3]:
                        CommandsPriority += (CommandTotal,)
                    else:
                        CommandsRegular += (CommandTotal,)
                else:
                    QUIT = True

            if QUIT:
                CommandsSorted = CommandsPriority + ("QUIT",)
            else:
                CommandsSorted = CommandsPriority + CommandsRegular

            CommandsTotal = ()

            # Process commands
            for CommandSorted in CommandsSorted: # (Method, URL-handler, Parameters, Priority)
                if CommandSorted == "QUIT":
                    xbmc.log("EMBY.emby.http: Async closed", 1) # LOGINFO
                    xbmc.log(f"EMBY.emby.emby: THREAD: ---<[ Async {self.EmbyServer.ServerData['ServerId']} ]", 0) # LOGDEBUG
                    return

                if not CommandSorted[3] and self.wait_for_main("ASYNC"):
                    return

                if self.socket_open_async("ASYNC"):
                    if utils.sleep(10):
                        xbmc.log(f"EMBY.emby.emby: THREAD: ---<[ Async {self.EmbyServer.ServerData['ServerId']} shutdown ]", 0) # LOGDEBUG
                        return

                    continue

                StatusCode, _, _ = self.socket_request(CommandSorted[0], CommandSorted[1], CommandSorted[2], False, 1, 1, "ASYNC", "", "")

                if StatusCode == 601: # quit
                    self.Queues["ASYNC"].clear()
                    self.socket_close("ASYNC")
                    xbmc.log(f"EMBY.emby.emby: THREAD: ---<[ Async {self.EmbyServer.ServerData['ServerId']} ] shutdown 2", 0) # LOGDEBUG
                    return

                if StatusCode in (600, 602, 603, 604, 605):
                    xbmc.log(f"EMBY.emby.http: Async retry {StatusCode}", 2) # LOGWARNING
                    self.socket_close("ASYNC")
                    continue

        self.Queues["ASYNC"].clear()
        self.socket_close("ASYNC")
        xbmc.log(f"EMBY.emby.emby: THREAD: ---<[ Async {self.EmbyServer.ServerData['ServerId']} ]", 0) # LOGDEBUG

    # Ping server -> keep http session open (timer)
    def Ping(self):
        xbmc.log(f"EMBY.emby.emby: THREAD: --->[ Ping {self.EmbyServer.ServerData['ServerId']} ]", 0) # LOGDEBUG

        while True:
            for _ in range(20): # ping every 10 seconds
                if utils.sleep(0.5):
                    xbmc.log(f"EMBY.emby.emby: THREAD: ---<[ Ping {self.EmbyServer.ServerData['ServerId']} shutdown or quit ]", 0) # LOGDEBUG
                    return

                if "MAIN" not in self.socket:
                    xbmc.log("EMBY.emby.http: Ping MAIN not running", 1) # LOGINFO
                    return

            if self.wait_for_main("PING"):
                return

            _, _, _ = self.request("POST", "System/Ping", {}, {}, True, "", False)

        xbmc.log(f"EMBY.emby.emby: THREAD: ---<[ Ping {self.EmbyServer.ServerData['ServerId']} ]", 0) # LOGDEBUG

    # Intros and Trailers
    def verify_intros(self, Intro):
        xbmc.log("EMBY.emby.http: THREAD: --->[ verify intros ]", 0) # LOGDEBUG

        if Intro['Path'].find("http") == -1: # Local Trailer
            Intro['Path'], _ = common.get_path_type_from_item(self.EmbyServer.ServerData['ServerId'], Intro, False, True)
            self.Intros.append(Intro)
            xbmc.log("EMBY.emby.http: THREAD: ---<[ verify intros ] valid local intro", 0) # LOGDEBUG
            return True

        status_code = self.EmbyServer.API.get_stream_statuscode(Intro['Id'], Intro['MediaSources'][0]['Id'])

        if status_code == 200:
            Intro['Path'], _ = common.get_path_type_from_item(self.EmbyServer.ServerData['ServerId'], Intro, False, True)
            self.Intros.append(Intro)
            xbmc.log("EMBY.emby.http: THREAD: ---<[ verify intros ] valid http", 0) # LOGDEBUG
            return True

        xbmc.log(f"EMBY.emby.http: Invalid Trailer: {Intro['Path']} / {status_code}", 3) # LOGERROR
        xbmc.log("EMBY.emby.http: THREAD: ---<[ verify intros ] invalid", 0) # LOGDEBUG
        return False

    def load_Trailers(self, EmbyId):
        ReceivedIntros = []
        self.Intros = []

        if utils.localTrailers:
            LocalTrailers = self.EmbyServer.API.get_local_trailers(EmbyId)

            for LocalTrailer in LocalTrailers:
                ReceivedIntros.append(LocalTrailer)

        if utils.Trailers:
            Intros = self.EmbyServer.API.get_intros(EmbyId)

            if 'Items' in Intros:
                for Intro in Intros['Items']:
                    ReceivedIntros.append(Intro)

        if ReceivedIntros:
            Index = 0

            for Index, Intro in enumerate(ReceivedIntros):
                if self.verify_intros(Intro):
                    break

            for Intro in ReceivedIntros[Index + 1:]:
                start_new_thread(self.verify_intros, (Intro,))

# Return empty data
def noData(StatusCode, Header, Binary):
    if Binary:
        return StatusCode, Header, b""

    return StatusCode, Header, {}
